# -*- coding: utf-8 -*-

"""
jsongit.constants
"""

import pygit2

GIT_SORT_NONE = pygit2.GIT_SORT_NONE
GIT_SORT_TOPOLOGICAL = pygit2.GIT_SORT_TOPOLOGICAL
GIT_SORT_TIME = pygit2.GIT_SORT_TIME
GIT_SORT_REVERSE = pygit2.GIT_SORT_REVERSE
